#!/usr/local/bin/perl
print "Welcome to the nandBot configure tool.\n\n";

print "This script will ask you a few simple questions to help configure nandBot.\n\n";

my $test = ask("Would you like to test your configuration to see if you can run nandBot on your computer?","y","y,n,yes=y,no=n");
print "\n";
if ( $test eq "y" ) {
    test(1,"use Net::IRC;print \"Net::IRC Okay\\n\";1") or exit 1;
    test(2,"use Time::Local;print \"Time::Local Okay\\n\";1") or exit 1;
}
else {
    print "Skipping...\n";
}

print "\nNow we need a little configuration information.\n";
my @config=();
my $nick = ask("What is this nandBot's nickname going to be?","nandBot");
push @config, "nick=$nick";
print "\n";

my $ownernick = ask("What is you (the owner)'s IRC nickname?");
print "\n";
my $owneremail = ask("What is your email address? This will be put in nandBot's realname field.\nPress enter if you don't want one:");
my $owner = $ownernick;
$owner .= "<$owneremail>" if $owneremail;
push @config, "owner=$owner";
print "\n";
print "Okay, nandBot's realname will be 'Owner: $owner'.\n\n\n";

print "Please choose an unimportant password to use for authenticating to nandBot.\n";
my $pass = ask("Password:");

print "\n\n\nNow this is the nasty part of the configuration.\nI need a list of rooms to auto-join on connect, and rooms I can talk in.\n\n";
print "Press enter when there are no more.\n";

while (1) {
    print "\n";
    my $r  = ask("Room:");
    last unless $r;
    my $aj = ask("Auto-Join: ","y","1,0,y=1,n=0,yes=1,no=0");
    my $ct = ask("Can I talk?","y","1,0,y=1,n=0,yes=1,no=0");
    push @config, "$r=$aj,$ct";
}

print "\n\n\nNow, who are my neighbors? I can consult my neighbors when someone asks me\nabout a factoid, and I'm clueless.\n\nMy neighbors must be other infobots, blootbots, or nandBots, or nasty problems\nmight surface.\n\nPress enter when there are no more. (Or if there are none at all).\n";

my @bots = ();
while (1) {
    my $b = ask("Neighbor's nick:");
    last unless $b;
    push @bots, $b;
}
push @config, "neighbors=".join(',',@bots);
print "\n\n\nWhat IRC servers can I connect to? I need at least one:\n";
while (1) {
    my $b = ask("Server:");
    last unless $b;
    push @servers, $b;
}

# It's showtime!
print "\n\n\n\nNow that I have all the information I need, I am ready to write\nthe nandBot configuration files.\n\n";
if ( ask("Would you like to review the configuration files before I write them?","y","y,n,yes=y,no=n") eq "y" ) {
    print join("\n  ","Here is the proposed nandBot.conf file:",@config);
    print "\nThe proposed data/$ownernick.users file will give you access level 10.\n";
    print join("\n  ","Here is the proposed serverlist file:",@servers);
    print "\nI will also create a few starting factoids.\n";
}
print "\n\n";
if ( ask("Last chance to abort! Write your new nandBot configuration?","y","y,n,yes=y,no=n") eq "y" ) {
    print "Writing nandbot.conf...";
    backup("nandbot.conf");
    open CONF, ">nandbot.conf";
    print CONF join("\n",@config);
    close CONF;
    print "done\n";

    print "Writing data/$ownernick.users...";
    backup("data/$ownernick.users");
    open CONF, ">data/$ownernick.users";
    print CONF "seenroom=configure.pl\n";
    print CONF "seentime=".time."\n";
    print CONF "seentxt=y\n";
    print CONF "level=10\n";
    close CONF;
    print "done\n";

    my @cra = (				# Array of random characters.
	       '.', '/', 0..9,
	       'A'..'Z', 'a'..'z');
    print "Writing data/$ownernick.pass...";
    backup("data/$ownernick.pass");
    open CONF, ">data/$ownernick.pass";
    print CONF crypt($pass,join('', @cra[rand 64, rand 64]));
    close CONF;
    print "done\n";

    print "Writing serverlist...";
    backup("serverlist");
    open CONF, ">serverlist";
    print CONF join("\n",@servers);
    close CONF;
    print "done\n";

    print "Writing data/nandbot.facts...";
    backup("data/nandbot.facts");
    open CONF, ">data/nandbot.facts";
    print CONF <<ENDFACT;
unsaved=0
locked=nandbot
creator=nandbot
forgot=
targetwait=#mercyground
content=<reply>I am a perl IRC robot with many features. See http://nandbot.sf.net/
author=nandbot
dunno=0
ENDFACT
    close CONF;
    print "done\n";

    print "\n\n\n Congratulations, you you have completed the configuration process!\nYou may now proceed with the next step in the installation instructions.\n";
}
else {
    print "Configuration unchanged.\n";
}
#############################################################################
sub ask {
    my ($q,$d,$c) = @_;
    $d ||= "";
  Retry:
    print "$q";
    print "\[$d\]" if $d;
    print " ";
    my $a = <STDIN>;
    $a =~ s/[\r\n]//g;
    $a = $d unless $a;
    if ( $c ) {
	$a = lc $a;
	if ( $a =~ /^q|quit$/ ) {
	    exit 0;
	}
	elsif ( $c !~ /(^|,)($a=([^,]+)(,|$)|$a(,|$))/ ) {
	    print "\nPlease try again\n";
	    goto Retry;
	}
	else {
	    $a = $3 if defined $3;
	}
    }
    return $a;
}
sub backup {
    my ($f) = @_;
    if ( -f "$f" ) {
	unlink "$f~" if -f "$f~";
	rename "$f","$f~";
	print "\nSaving existing $f as $f~\n";
    }
}
sub test {
    my ($no,$testcode) = @_;
    my $r = eval $testcode;
    my $s=1;
    ($s=0,warn "Error during test $no: $!\n") if !defined $r;
    ($s=0,warn "Error during test $no: $@\n") if $@;
    ($s=0,warn "Error during test $no\n") if !$r;
    return $s;
}
