#!/usr/bin/perl
#
#  nandBot
#     IRC Robot
#
BEGIN {print "Initalizing..."; unshift @INC, "."}
our $version = 0.50;
my $nbstr = "nandBot $version: http://nandbot.sf.net/";

=head1 NAME

nandBot - IRC Robot

=head1 SYNOPSIS

nandbot.pl [server]

=head1 OPTIONS

=over 4

=item server

IRC Server to connect to. nandBot will try to connect to this server before trying each server in the server list.

=back

=head1 EXAMPLES

=over 4

=item nandbot.pl

Normal run.

=item nandbot.pl irc.fusionchat.net

Use irc.fusionchat.net as the default server.

=back

=head1 PERL MODULES

=over 4

=item Net::IRC

=item Time::Local

=item Net::Dict (optional)

=item Net::Google (optional)

=back

=head1 AUTHOR

nand <nandhp@users.sourceforge.net>

=head1 SEE ALSO

perl(1), NET::IRC, Time::Local, Net::Dict, Net::Google, RFC1459,
irchelp.org, http://nandbot.sf.net/

=cut

#=item nandbot.pl irc.fusionchat.net 1
#
#Use irc.fusionchat.net as the default server (Debugging mode).
#

# use strict;
use Net::IRC;
use Time::Local;
use warnings;
#use GetOpt::Long;
#use strict;

#
#  Create the IRC and Connection objects
#

$| = 1;

my $botoff = 0;
my $offreason = "";
my $logfile = "data/log_file";
my $mc=0;

my $knock = "";

our @commands = ();

#my @logs = ();				# Deprecated. See the nblog function.

my $irc = new Net::IRC;

print "Creating connection to IRC server...\n";

# This code finds a server to connect to
my @servers = (
	       shift(@ARGV),
	       "",
);

#Bots to ask for help.
my @neighbors = ();
my %neighborstats = ();
my %isneighbor = ();			# So we don't get as many endless loops
my @words = (				# Random words
	     "onion",
	     "tableknife",
	     "robot",
	     "trout",
	    );
my @cra = (				# Array of random characters.
	   '.', '/', 0..9,
	   'A'..'Z', 'a'..'z');
my @ateball = (				# 8ball responses
	       "Yes",
	       "Of Course!",
	       "Why should I tell you?",
	       "I don't feel like answering.",
	       "Go away.",
	       "The things I get asked sometimes. Would you belive 'What brand of soap do you use?'?",
	       "And you're asking me for what reason?",
	       "Asking difficult questions is so sphinx-like... Will you eat me if I'm wrong?",
	       "I could tell you, but then I'd have to /KILL you.",
	       "You don't want to know.",
	       "The answer is the same as 75/0.",
	       "The answer is the same as the square root of -72.",
	       "Swim to latitude 50 north, longitude 40 west (that is magic) for the answer.",
	       "To find THAT out you must answer me this: What has but one name, and in the morning goes on four feet, in the afternoon on two, and in the evening on three?",
	       "That is classified information.",
	       "Only time will tell.",
	       "Do I look like I know?",
	       "The answer is somewhere... over the rainbow...",
	      );
# Countdown feature
my @dates = (				# DD/(MM-1)
	     [2,1],			# Groundhog Day
	     [14,1],			# Valentine's Day
	     [17,2],			# St. Patrick's Day
	     #[1,4],			# nandBot's Birthday
	     [4,6],			# July 4th
	     [31,9],			# Halloween
	     #[15,10,2002],		# Chamber Of Secrets Release
	     [27,10,2003],		# Thanksgiving Day
	     [25,11],			# Christmas
	     [31,11],			# New Year's Eve
	    );
my @datenames = (
		 "Groundhog Day",
		 "Valentine's Day",
		 "St. Patrick's Day",
		 #"nandBot's Birthday",
		 "Independence Day",
		 "Halloween",
		 #"Chamber Of Secrets Release",
		 "Thanksgiving Day",
		 "Christmas",
		 "New Year's Eve",
		);
my @ccquotes = ("How do I click a mouse?",
		"IQ: Cockroach, Krill, Pond Scum, lkay",
		"Help me where s the letter eye on my keyboard?",
		"Whew it took me a couple of centuries but I _did_ learn how to kick someone.",
	       );
my @msquotes = ("Coffee will not be shipped with WindowsXP.",
		"Outlook bug count: 9876 - % of code scanned: 5%",
		"You can't hack windows: There are too many Windows and Gates",
	       );
# Prefixes for "x is y"
my @prefixes = (
		"Rumor has it, ","Well I've heard ", "According to rumor, ",
		"Somebody said ","In my evesdropping days I heard ",
	       );
my @secret = ('M;F%N9$)O="!W;W5L9"!L:6ME(\'1O(&5X=&5N9"!A(\'-P96-I86P@=&AA;FMS%(\'1O.@H`',
	      'M675U:6-H:2`M($9U<VEO;D-H870@;F5T861M:6XL(&YA;F1";W0@:&]S="X*',
	      'M4V]U<F-E1F]R9V4@+2!&;W(@:&]S=&EN9R!T:&4@;F%N9$)O="!P<F]J96-T"+@H`',
	      'K4VQA<VAD;W0@+2!&;W(@82!S=&5A9\'D@9F5E9"!O9B!G;V]D(&YE=W,N"@``',
	      '806YD(&QA<W0@8G5T(&YO="!L96%S=#H*',
	      '6("T@1F]R(\'5S:6YG(&YA;F1";W0N"@``\');',
	     );
our %users    = ();		  	# User Information (seen, IRCPoints)
our %general  = ();			# General Information (maxpoints, etc)
our %factoids = ();			# Factoids
our %userauth = ();			# Authentication of users.
our %rooms    = ();			# Room information

if ( !-d "data/" ) {
    print "Creating data directory.\n";
    mkdir "data";
}
if ( !-d "logs/" ) {
    print "Creating logs directory.\n";
    mkdir "logs";
}

opendir DATADIR, "data/";
my @datadir=readdir DATADIR;
my @datafile;
my @dataline;
my $usrfile;

#Read saved information about users
shift @datadir; # Get rid of
shift @datadir; # . and ..
foreach (@datadir) {
    if ( $_ =~ /.users$/ ) {
	$usrfile=$_;
	open USERFILE, "data/" . $usrfile;
	for (my $i=0;$i<6;$i++) {
	    chop $usrfile;
	}
	@datafile=<USERFILE>;
	foreach (@datafile) {
	    $_ =~ s/[\r\n]//g;
	    @dataline = split "=", $_, 2;
	    $dataline[0] =~ s/[\r\n]//g;
	    $dataline[1] =~ s/[\r\n]//g;
	    $users{lc $usrfile}{lc $dataline[0]}=$dataline[1];
	    $users{lc $usrfile}{unsaved}=0;
	}
	close USERFILE;
    }
    if ( $_ =~ /.facts$/ ) {
	$usrfile=$_;
	open USERFILE, "data/" . $usrfile;
	for (my $i=0;$i<6;$i++) {
	    chop $usrfile;
	}
	@datafile=<USERFILE>;
	foreach (@datafile) {
	    $_ =~ s/[\r\n]//g;
	    @dataline = split "=", $_, 2;
	    $dataline[0] =~ s/[\r\n]//g;
	    $dataline[1] =~ s/[\r\n]//g;
	    $factoids{lc $usrfile}{lc $dataline[0]}=$dataline[1];
	}
	close USERFILE;
    }
    if ( $_ eq "general" ) {
	open USERFILE, "data/general";
	@datafile=<USERFILE>;
	foreach (@datafile) {
	    $_ =~ s/[\r\n]//g;
	    @dataline = split "=", $_, 2;
	    $general{lc $dataline[0]}=$dataline[1];
	}
	close USERFILE;
    }
}
closedir DATADIR;

# Config file
open USERFILE, "nandbot.conf";
@datafile=<USERFILE>;
foreach (@datafile) {
    $_ =~ s/[\r\n]//g;
    next unless $_;			# blank lines
    @dataline = split "=", $_, 2;
    $dataline[0]=lc $dataline[0];
    if ( $dataline[0] =~ /^[\#\&]/ ) {
	$dataline[1] =~ /^(\d)\,(\d)$/;
	$rooms{$dataline[0]}{canTalk} = $2;
	$rooms{$dataline[0]}{autoJoin} = $1;
    }
    elsif ( $dataline[0] =~ /^neighbou?rs$/ ) {
	foreach ( split ',', $dataline[1] ) {
	    push @neighbors, $_;
	}
    }
    else {
	$general{lc $dataline[0]}=$dataline[1];
    }
}
close USERFILE;

sub slerror {
    warn "Error opening server list: $!\n";
    goto slskip;
}
open SLIST, "serverlist" or slerror();
my @rilist = <SLIST>;
for ($i=0;$i<=$#rilist;$i++ ) {
    $rilist[$i] =~ s/[\r\n]//g;
}
push @servers, @rilist;

slskip:

my $conn;
my $connected;

foreach (@servers) {
    next unless $_;
    print "Trying $_...\n";
    my ($server,$foo,$port) = m/^([^\:]+)(\:(\d+))?$/;
    $port ||= 6667;
    $conn = $irc->newconn(Server   => $server,
			  Port	   => $port,
			  Nick	   => $general{nick},
			  Ircname  => 'Owner: '.$general{owner},
			  Username => 'nandBot') or next;
    $connected = 1;
    last;
}
die "Could not connect" unless $connected;

# Authenticate to NickServ
sub nickserv {
    my $self = $_[0];
    open NS, ".nickserv";
    my $p=<NS>;
    close NS;
    $p =~ s/[\r\n]//g;
    print "IDing to NickServ: " . "*"x length($p) ."\n";
    $self->privmsg('NickServ','IDENTIFY '.$p);
    $p = "*"x length($p);		# Very overkill
}
#
#  Here are the handler subroutines. Fascinating, huh?
#

# What to do when the bot successfully connects.
sub on_connect {
    my $self = shift;
    my $mynick = $self->nick;
    $self->mode($mynick,"+BG");
    nickserv($self);
    foreach (keys %rooms) {
	if ( $rooms{$_}{autoJoin} ) {
	    print "Joining $_\n";
	    $self->join($_);
	}
    }
}

# Used for logging..
sub dd {				# Return the number in double-didgets.
    my $num = $_[0];			# Yes, I can use sprintf, but I didn't
    if ( $num < 10 ) {			# know that at the time.
	return "0$num";
    }
    else {
	return "$num";
    }
}
sub nblog {				# I don't log everything because then
    ($channel,$entry) = @_;		# the log file would be huge!

    if ( $channel =~ /^[^#]/ || $channel =~ /^#[^ ]* \[OR\]$/i ) {
	# It's a whisper or overided message.
	$channel =~ s/ \[OR\]$//i;
	$logit = 1;
    }
    else {
	$logit = 0;
    }
    if ( $logit == 1 ) {
	($sec,$min,$hour,$ddd,$mm,$yr) = localtime(time);
	$yr+=1900;
	$mm += 1;
	if ( $hour >= 12 ) {
	    $hour -= 12;
	    $pm = 1;
	}
	else {
	    $pm = 0;
	}
	$ts  = join "-",(dd($mm),dd($ddd));
	$ts .= " ". join ":",(dd($hour),dd($min),dd($sec));
	$ts .= $pm ? "P" : "A";
	$entry =~ s/[\r\n]//g;
	#push @logs, "[$ts] $entry";	# Now I just write it to today's file
	open LOG, ">>logs/".dd($mm).dd($ddd).$yr;
	print LOG "[$ts] $entry\n";
	close LOG;
	#open LOGFILE, ">>$logfile";
	#print LOGFILE "[$ts] $entry\n";
	print "[$ts] $entry\n";
	#close LOGFILE;
    }
}

# Handles some messages you get when you connect
sub on_init {
    my ($self, $event) = @_;
    my (@args) = ($event->args);
    shift (@args);

    print "*** @args\n";
}

# What to do when someone leaves a channel the bot is on.
sub on_part {
    my ($self, $event) = @_;
    my ($channel) = ($event->to)[0];

    nblog($channel, sprintf("*** %s has left channel %s\n", $event->nick, $channel));
}
sub on_kick {
    my ($self, $event) = @_;
    my ($channel) = ($event->args)[0];
    my ($nick, $mynick) = ($event->nick, $self->nick);

    nblog($channel,sprintf("*** %s has been kicked from channel %s by %s: %s\n", $event->to, $channel, $event->nick ,($event->args)[1]));

    if ( ($event->to)[0] eq $mynick ) {
	$self->join($channel);
	$self->privmsg($event->nick, "Don't kick me!");
    }
    my $foo = ($event->to)[0];
    if ( ($users{lc $foo}{level}||0) >= 10 ) {
	$self->privmsg($event->nick, "Don't kick nandBot Administrators!");
    }
    else {
	my @to = $event->to;
	$users{$to[0]}{kickstats}++;
	#print "$to[0] has $users{$to[0]}{kickstats} kicks on record\n";
    }
}

# What to do when someone joins a channel the bot is on.
sub on_join {
    my ($self, $event) = @_;
    my ($channel) = ($event->to)[0];
    my $from = $event->from;
    nblog($channel, sprintf("*** %s has joined channel %s\n",
			  $event->nick, $channel));

    if ((( $users{lc $event->nick}{level} || 0 ) >= 10)&&($userauth{$nick} eq $from) ) { 
	$self->mode($channel, "+o", $event->nick);
    }
    elsif ((( $users{lc $event->nick}{level} || 0 ) >= 5)&&($userauth{$nick} eq $from) ) { 
	$self->mode($channel, "+h", $event->nick);
    }
}

# Handle mode changes
sub on_mode {
    my ($self, $event) = @_;
    my ($channel) = ($event->to);

    nblog($channel,sprintf("*** %s on %s has set modes: %s\n", $event->nick, $channel, join(' ',($event->args))));
}
sub on_invite {
    my ($self, $event) = @_;

    $self->join(($event->args)[0]);	# Join the room
}

# A string of how long ago something happened. Takes one CTime argument.
sub since {
    my $sincewhen = $_[0];
    my $now = time;
    my $sec_since = 0;
    my @format = (
		  0,	#days	- 86400
		  0,	#hours	- 3600
		  0,	#min	- 60
		  0	#sec	- 1
		 );

    $now =~ s/[\r\n]//g;

    $sec_since = $now - $sincewhen;
    # And in case you want until too...
    $sec_since = $sincewhen - $now if $now < $sincewhen;

    while ($sec_since >= 86400) {	# see how many days
	$format[0]++;
	$sec_since -= 86400;
    }
    while ($sec_since >= 3600) {	# see how many hours
	$format[1]++;
	$sec_since -= 3600;
    }
    while ($sec_since >= 60) {		# see how many min
	$format[2]++;
	$sec_since -= 60;
    }
    $format[3] = $sec_since;

    my $f_str = "";
    $f_str .= "$format[0] day".($format[0]>1?"s ":" ") if $format[0] > 0;
    $f_str .= "$format[1] hour".($format[1]>1?"s ":" ") if $format[1] > 0;
    $f_str .= "$format[2] minute".($format[2]>1?"s ":" ") if $format[2] > 0;
    $f_str .= "and $format[3] second".($format[3]>1?"s ":" ") if $format[3] > 0;
    chop $f_str if $f_str =~ / $/;
    $f_str = "0 seconds" if $f_str eq "";
    return $f_str;
}

# Find out if nandBot can talk in a room, and if there is a room to talk in
sub cantalk {
    # cantalk($room);
    if ( $rooms{lc $_[0]}{canTalk} ) {
	return 1;
    }
    else {
	return 0;
    }
}

# Save information about users
sub save {
    print "Saving...";
#    print "log...";
#    open LOGFILE, ">>$logfile" or warn "Cannot open logfile!";
#    foreach( @logs ) {
#	print LOGFILE "$_\n";
#    }
#    @logs = ();
#    close LOGFILE;
    print "user info...";
    foreach (keys %users) {
	my $tmp=$users{lc $_};
	next if !$users{lc $_}{unsaved};
	open USERFILE, ">data/" . $_ . ".users";
	$users{lc $_}{unsaved}=0;
	my @keys = keys %{$tmp};
	my @values = values %{$tmp};
	foreach (@keys) {
	    print USERFILE $_,"=",$tmp->{$_},"\n";
	}
    }
    print "factoids...";
    foreach (keys %factoids) {
	my $tmp=$factoids{lc $_};
	next if !$factoids{lc $_}{unsaved};
	open USERFILE, ">data/" . $_ . ".facts" or warn "Could not open nandBot factiod data file for factoid $_: $!";
	$factoids{lc $_}{unsaved}=0;
	my @keys = keys %{$tmp};
	my @values = values %{$tmp};
	foreach (@keys) {
	    print USERFILE $_,"=",$tmp->{$_},"\n";
	}
    }
    print "other info...";
    open USERFILE, ">data/general";
    foreach ( keys %general) {
	print USERFILE $_,"=",$general{$_},"\n";
    }
    close USERFILE;
    $mc=0;
    print "done\n";
}

# Parse a nandBot command. This is where all the exciting stuff happens :)
sub nandbotcmd {
    my $self = $_[0];
    my $args = $_[1];
    my $nick = $_[2];
    my $isroom = $_[3];
    my $room = $_[4];
    my $chan = $room;
    my $wap = $_[5];
    my $from = $_[6];#$userauth{$nick};
    # Fun, eh?

    #if ( $isroom) {
    #	$self->privmsg($room, $args); # appears in the message.
    #}
    #else {
    #$self->privmsg($isroom ? $room : $nick, $args); # appears in the message.
    #}
    if ( $botoff && ( $users{lc $nick}{level} || 0 ) < 10) {
	$self->privmsg($nick,"$offreason");
	return;
    }
    if ( $botoff && ( $users{lc $nick}{level} || 0 ) >= 10 && $args !~ /^on$/i) {
	$self->privmsg($nick,"WARNING nandBot is off: $offreason");
    }
    my $adminexec = 0;
#    if ( $userauth{$nick} != $from ) {
	#host is wrong
#    }
    #print "UA: $userauth{$nick} FROM $from\n";
    if ( ( $users{lc $nick}{level} || 0 ) >= 10 && !$wap #) {
	 && $userauth{$nick} eq $from
       ) { # make sure host is right
	if ( $args =~ /^quit( (.+))?$/i ) {
	    #save();
	    my $qm = "Quit command received from $nick";
	    $qm .= ": $2" if $2;
	    $self->quit($qm);
	    quit();
	    exit 0;
	}
	if ( $args =~ /^takepoints (.+) (.+)$/i ) {
	    my $pointsto=$users{lc $1}{doredir} ? $users{lc $1}{redirect} : $1;
	    $users{lc $pointsto}{points} -= $2;
	    $users{lc $pointsto}{unsaved} = 1;
	    $adminexec = 1;
	}
	if ( $args =~ /^takepoints_nr (.+) (.+)$/i ) {
	    my $pointsto= $1;
	    $users{lc $pointsto}{points} -= $2;
	    $users{lc $pointsto}{unsaved} = 1;
	    $adminexec = 1;
	}
	if ( $args =~ /^delpoints (.+)$/i ) {
	    my $pointsto=$users{lc $1}{doredir} ? $users{lc $1}{redirect} : $1;
	    $users{lc $1}{points} = 0;
	    $users{lc $1}{unsaved} = 1;
	    $adminexec = 1;
	}
	if ( $args =~ /^delpoints_nr (.+)$/i ) {
	    my $pointsto= $1;
	    $users{lc $1}{points} = 0;
	    $users{lc $1}{unsaved} = 1;
	    $adminexec = 1;
	}
	if ( $args =~ /^givepoints (.+) (.+)$/i ) {
	    my $pointsto=$users{lc $1}{doredir} ? $users{lc $1}{redirect} : $1;
	    $users{lc $1}{points} += $2;
	    $users{lc $1}{unsaved} = 1;
	    $adminexec = 1;
	}
	if ( $args =~ /^movepoints (.+) (.+) (.+)$/i ) {
	    my $pointsto=$users{lc $1}{doredir} ? $users{lc $1}{redirect} : $1;
	    my $pointsfr=$users{lc $2}{doredir} ? $users{lc $2}{redirect} : $2;
	    $users{lc $pointsto}{points} -= $3;
	    $users{lc $pointsfr}{points} += $3;
	    $users{lc $pointsto}{unsaved} = 1;
	    $users{lc $pointsfr}{unsaved} = 1;
	    $adminexec = 1;
	}
	if ( $args =~ /^getencpass (.+)$/i ) {
	    open PASSFILE, "data/". $1 .".pass";
	    $self->privmsg( $nick, (<PASSFILE>||"Error: No password for $1"));
	    close PASSFILE;
	    $adminexec = 1;
	}
	if ( $args =~ /^access (.+) (\d+)$/i ) {
	    if ( -f "data/$1".".pass" ) {
		$users{lc $1}{level} = $2;
		$users{lc $1}{unsaved} = 1;
		$self->privmsg( cantalk($room) ? $room : $nick,"$1's access level is now $2");
		#$self->privmsg( $1,"$nick has given you nandBot access level $2. If online, now you know. If not, wrong number");
	    }
	    else {
		$self->privmsg( cantalk($room) ? $room : $nick,"Error: User must create password");
	    }
	    $adminexec = 1;
	}
	if ( $args =~ /^off$/i ) {
	    $botoff=1;
	    $offreason="nandBot is unavailable at this time. Please try again in a few minutes.";
	    $adminexec = 1;
	}
	if ( $args =~ /^off (.+)$/i ) {
	    $botoff=1;
	    $offreason="nandBot is unavailable at this time: $1";
	    $adminexec = 1;
	}
	if ( $args =~ /^on$/i ) {
	    $botoff=0;
	    $adminexec = 1;
	}
	if ( $args =~ /^umode ([\-\+\w]+)( (.+))?$/i) {
	    $self->mode($self->nick,$1,$3);
	    $adminexec = 1;
	}
    }
    if ( ($users{lc $nick}{level} || 0) >= 5 && !$wap
	&& $userauth{$nick} eq $from
       ) { # make sure host is right
	if ( $args =~ /^join (.+)/i ) {
	    $self->join($1);
	    $adminexec = 1;
	}
	if ( $args =~ /^part (.+)/i ) {
	    $self->part($1);
	    $adminexec = 1;
	}
	elsif ( $args =~ /^part/i && $isroom ) { #this is elsif because
	    $self->part($room);		         #of the other part
	    $adminexec = 1;
	}
	if ( $args =~ /^kick ([^ ]+) (.+)$/i ) {
	    if ( $users{lc $1}{level} > $users{lc $nick}{level} ) {
		$self->privmsg(cantalk($room) ? $room : $nick,"You are not allowed to kick that user.");
	    }
	    else {
		$self->kick($2,$1,"Kick command used by $nick");
	    }
	    $adminexec = 1;
	}
	elsif ( $args =~ /^kick (\w+)$/i && $isroom) {
	    if ( $users{lc $1}{level} > $users{lc $nick}{level} ) {
		$self->privmsg(cantalk($room) ? $room : $nick,"You are not allowed to kick that user.");
	    }
	    else {
		$self->kick($room,$1,"Kick command used by $nick");
	    }
	    $adminexec = 1;
   	}
	if ( $args =~ /^mode (#\w+) ([\-\+\w]+)( (.+))?$/i) {
	    $self->mode($1,$2,$4);
	    $adminexec = 1;
	}
	elsif ( $args =~ /^mode ([\-\+\w]+)( (.+))?$/i && $isroom) {
	    $self->mode($room,$1,$3);
	    $adminexec = 1;
	}
	if ($args =~ /^wantnick( (.+))?$/i) {
	    $self->nick($1 ? $1 : "nandBot");
	    nickserv($self) unless $1;
	    $adminexec = 1;
        }
 	if ( $args =~ /^say ([^ ]+) (.+)$/i && !$isroom ) {
	    $self->privmsg($1,$2);
	    $adminexec = 1;
	}
	if ( $args =~ /^me ([^ ]+) (.+)$/i  && !$isroom ) {
	    $self->me($1,$2) ;
	    $adminexec = 1;
	}
	if ( $args =~ /^say (.+)$/i && $isroom) {
	    $self->privmsg($room,$1);
	    $adminexec = 1;
	}
	if ( $args =~ /^me (.+)$/i && $isroom) {
	    $self->me($room,$1);
	    $adminexec = 1;
	}
	if ( $args =~ /^save$/i ) {
	    save();
	    $adminexec = 1;
	}
    }
    if ( (( $users{lc $nick}{level} || 0 ) >= 5)
	 && ( $wap || ($userauth{$nick} ne $from) )) {
	$self->privmsg($nick,"You must authenticate to nandBot by whispering the AUTH command to me.");
	#$adminexec = 1
    }
    return if $adminexec == 1;
    if ( $args =~ /^help$/i ) {
   	$self->privmsg(cantalk($room) ? $room : $nick,"For help, please visit http://nandbot.sf.net/commands.shtml");
    }
    elsif ( $args =~ /^adminhelp$/i ) {
	$self->privmsg(cantalk($room) ? $room : $nick,"For help, please visit http://nandbot.sf.net/admincmds.shtml");
    }
    elsif ($args =~ /^optout anger$/i ) {
	$users{lc $nick}{angeroptout}=2;
	$self->privmsg($nick,"Success! You can turn this feature back on with the command: 'OPTIN ANGER'");
    }
    elsif ($args =~ /^optin anger$/i ) {
	$users{lc $nick}{angeroptout}=1;
	$self->privmsg($nick,"Success!");
    }
    elsif ($args =~ /^8ball (.+)$/i ) {
	$self->privmsg(cantalk($room) ? $room : $nick,ateball());
    }
    elsif ( $args =~ /^lock (.+)$/i ) {
	$args =~ s/\\is/is/g;
	if ( $nick eq $factoids{lc $1}{creator} || $nick eq $factoids{lc $1}{author}) {
	    if ( $factoids{lc $1}{locked} eq "notlocked") {
		$factoids{lc $1}{locked} = $nick;
		$factoids{lc $1}{unsaved} = 1;
		$self->privmsg(cantalk($room) ? $room : $nick,"Factoid Locked");
	    }
	    else {
		$self->privmsg(cantalk($room) ? $room : $nick,"Factoid is already locked ($factoids{lc $1}{locked} locked it).");
	    }
	}
	elsif ( $nick ne $factoids{lc $1}{creator} && $nick ne $factoids{lc $1}{author} ) {
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid not owned by you");
	}
	elsif ( $factoids{lc $1}{locked} ne "notlocked" ) {
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid is locked");
	}
	else {
	    $self->privmsg(cantalk($room) ? $room : $nick,"Cannot lock Factoid");
	}
    }
    elsif ( $args =~ /^unlock (.+)$/i ) {
	$args =~ s/\\is/is/g;
	if ( $nick eq $factoids{lc $1}{locked} || $nick eq $factoids{lc $1}{creator}){
	    $factoids{lc $1}{locked} = "notlocked";
	    $factoids{lc $1}{unsaved} = 1;
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid Unlocked");
	}
	elsif ( $nick ne $factoids{lc $1}{locked} ) {
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid not locked by you");
	}
	elsif ( $factoids{lc $1}{locked} eq "notlocked" ) {
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid is already unlocked");
	}
	else {
	    $self->privmsg(cantalk($room) ? $room : $nick,"Cannot unlock Factoid");
	}
    }
    elsif ( $args =~ /^forget (.+)$/i ) {
	$args =~ s/\\is/is/g;
	if ( !exists($factoids{lc $1}) ) {
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid does not exist");
	}
	else {
	    if ( ( $factoids{lc $1}{locked} eq "notlocked" || ($users{$nick}{level} || 0) >= 10 ) && exists($factoids{lc $1})) {
		delete $factoids{lc $1};
		unlink "data/".lc $1.".facts";
		$self->privmsg(cantalk($room) ? $room : $nick,"Factoid has been forgotten");
	    }
	    elsif ( $factoids{lc $1}{locked} ne "notlocked" ) {
		$self->privmsg(cantalk($room) ? $room : $nick,"Factoid is locked");
	    }
	    else {
		$self->privmsg(cantalk($room) ? $room : $nick,"Cannot forget Factoid");
	    }
	}
    }
    elsif ( $args =~ /^points (.+)$/i || $args =~ /^stats (.+)$/i ) {
	$self->privmsg(cantalk($room) ? $room : $nick, "$1 has " . ( $users{lc $1}{points} || 0 ) . " IRCPoints");
    }
    elsif ( $args =~ /^rot13 (.+)$/i ) {
	my $rm = $1;
	$rm =~ tr/a-zA-Z/n-za-mN-ZA-M/;
	$self->privmsg(cantalk($room) ? $room : $nick, "$rm");
    }
    elsif ( $args =~ /^maxpoints$/i ) {
	$self->privmsg(cantalk($room) ? $room : $nick, "The person with the most IRCPoints is $general{maxpointsnick}. $general{maxpointsnick} has $general{maxpoints} IRCPoints.");
    }
    elsif ( $args =~ /^paypoints (.+) (.+)$/i ) {
	if ( $2 > $users{lc $nick}{points} ) {
	    $self->privmsg($nick, "Sorry, you don't have that many points.");
	    return 0;
	}
	if ( $2 < 0 ) {
	    $self->privmsg($nick, "Sorry, that's a negative number.");
	    return 0;
	}

	my $sendto=$users{lc $1}{doredir} ? $users{lc $1}{redirect} : $1;
	$users{lc $nick}{points} -= $2;
	$users{lc $sendto}{points} += $2;
	$factoids{lc $nick}{unsaved} = 1;
	$factoids{lc $sendto}{unsaved} = 1;
	$self->privmsg($1,"$nick paid you $2 IRCPoints. Resent to $sendto") if $1 ne $sendto;
	$self->privmsg($sendto, "You got IRCPoints!");
	$self->privmsg($sendto, "$nick has paid you $2 IRCPoints.");
	$self->privmsg($nick, "Transaction Complete");
    }
    elsif ( $args =~ /^getpoints (.+) (.+)$/i ) {
	$self->privmsg($1, "$nick wishes to be paid $2 IRCPoints.");
	$self->privmsg($1, "To pay $nick whisper to me 'paypoints $nick $2'.");
	$self->privmsg($nick, "Awaiting Authorization");
    }
    elsif ( $args =~ /^redir$/i ) {
	$users{lc $nick}{doredir}=0;
	$self->privmsg($nick,"Redirection OFF");
    }
    elsif ( $args =~ /^redir (.+)$/i ) {
	$users{lc $nick}{doredir}=1;
	$users{lc $nick}{redirect}=$1;
	$self->privmsg($nick,"Redirection ON. Redirecting to $1. Your IRCPoints have been moved to this nick. You can use paypoints to move your IRCPoints.");
    }
    elsif ( $args =~ /^clearkicks( ([^ ]+))?$/i ) {
	my $cnick = $nick;
	if ( ($users{lc $nick}{level} || 0) >= 5 && $2 ) {
	    $cnick = $2;
	}
	$users{$cnick}{kickstats} = 0;
	$self->privmsg($nick,"Kickstats for $cnick cleared.");
    }
    elsif ( $args =~ /^seen ([^ ]+)$/i ) {
	my $sstr = since($users{lc $1}{seentime})
	  . " ago in "
	    . $users{lc $1}{seenroom} 
	      . " saying "
		. $users{lc $1}{seentxt};
	my $kstr = "";
	#$kstr = ". $1 Has $users{lc $1}{kickstats} kicks on record."
	#  if $users{lc $1}{kickstats};
	if ( $1 =~ /^nandBot$/i ) {
	    $self->privmsg(cantalk($room) ? $room : $nick, "Here! Here!");
	}
	elsif ( $1 =~ /^$nick$/i ) {
	    $self->privmsg(cantalk($room) ? $room : $nick, "That would be you...?");
	}
#	elsif ( $1 =~ /^nand$/i ) {
#	    $self->privmsg(cantalk($room) ? $room : $nick, "My master was last seen at $sstr");
#	}
	elsif ( defined($users{lc $1}{seenroom}) ) {
	    $self->privmsg(cantalk($room) ? $room : $nick, "I last saw $1 at $sstr". $kstr);
	}
	else {
	    $self->privmsg(cantalk($room) ? $room : $nick, "Sorry, I haven't seen $1");
	}
    }
    elsif ( $args =~ /^factinfo (.+)$/i ) {
	$args =~ s/\\is/is/g;
	if ($factoids{lc $1}{content}) {
	    my $infostr = "$1 -- Created by: $factoids{lc $1}{creator}. Last modified by: $factoids{lc $1}{author}. ";
	    if ($factoids{lc $1}{locked} ne "notlocked" ) {
		$infostr .= "Factoid has been locked by $factoids{lc $1}{locked}."
	    }
	    else {
		$infostr .= "Factoid is not locked."
	    }
	    $self->privmsg(cantalk($room) ? $room : $nick ,"$infostr");
	}
	else {
	    $self->privmsg(cantalk($room) ? $room : $nick, "Factoid $1 does not exist.");
	}
    }
    elsif ( $args =~ /^countdown$/i ) {
	my @curtime = localtime;
	my $curmon = $curtime[4];
	my $curmday= $curtime[3];
	my $curyr  = $curtime[5];
	my $destime;
	my $i = 0;
	foreach (@dates) {
	    if (($_->[2] || $curyr+1900) == $curyr+1900) {
		if ( $_->[1] > $curmon ||
		     ($_->[0] > $curmday && $_->[1] >= $curmon )) {
		    $destime = timelocal(0,0,0,$_->[0],$_->[1],$curyr);
		    last;
		}
	    }
	    $i++;
	}
	$self->privmsg(cantalk($room) ? $room : $nick,
		       "Time until $datenames[$i]: ".since($destime));
    }
    elsif ( $args =~ /^setpass (.+)$/i ) {
	if ( $userauth{$nick} eq $from || !(-f "data/$nick.pass")) {
	    open PASSFILE, ">data/$nick.pass";
	    print PASSFILE crypt($1,join('', @cra[rand 64, rand 64]));
	    close PASSFILE;
	    $self->privmsg($nick,"Password set.");
	}
	else {
	    $self->privmsg($nick,"You must authenticate to nandBot by whispering the AUTH command to me. If you have forgotton your password, you may send a memo to nand.");
	}
    }
    elsif ( $args =~ /^status$/i ) {
	#print "\n\nStatus:\n";
	my $uptime = since($general{startup}); # How long since startup
	my $statstr = "I have been running $uptime. In this session, there have been $general{commands} commands, $general{mods} modifications".($general{lastmod} ? ", most recently $general{lastmod}, ":" ")."and $general{messages} messages to rooms ($general{via} via another robot). I am running on $^O.";
	my $infostr = "I reference ".scalar(keys(%factoids))." factoids, and know about ".scalar(keys(%users))." users.";
	#print "| $statstr\n| $infostr\n\n";
	$self->privmsg(cantalk($room) ? $room : $nick, "$statstr");
	$self->privmsg(cantalk($room) ? $room : $nick, "$infostr");
    }
    elsif ( $args =~ /^cookie$/i ) {
	my @keys = keys %factoids;
	my $fn = $keys[rand(scalar(@keys))];
	my $fc = $factoids{$fn}{content};
	#print $fc;
	$self->privmsg(cantalk($room) ? $room : $nick, "$fn is $fc");
    }
    elsif (($args !~ /^what is /i) && ($args =~ /^(.+?) \\{0}is also (.+)$/i)) {
	$args =~ s/\\is/is/g;
	if ( ($factoids{lc $1}{locked} || "notlocked") eq "notlocked" ) {
	    $factoids{lc $1}{content} .= " and $2";
	    $factoids{lc $1}{creator} = $nick unless defined($factoids{lc $1}{creator});
	    $factoids{lc $1}{author} = $nick;
	    $factoids{lc $1}{locked} = "notlocked";
	    $factoids{lc $1}{forgot} = "";
	    $factoids{lc $1}{unsaved} = 1;
	    $general{mods}++;
	    $general{lastmod}=lc $1;
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid changed");
	}
	else {
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid is locked");
	}
    }
    elsif ( ($args !~ /^(?:what|who|where) (?:is|was) /i) && ($args =~ /^(.+?) \\{0}is (.+)$/i) ) {
	$args =~ s/\\is/is/g;
	if ( ($factoids{lc $1}{locked} || "notlocked") eq "notlocked" ) {
	    $factoids{lc $1}{content} = $2;
	    my $changed = defined($factoids{lc $1}{creator}) ? "changed" : "created";
	    $factoids{lc $1}{creator} = $nick unless defined($factoids{lc $1}{creator});
	    $factoids{lc $1}{author} = $nick;
	    $factoids{lc $1}{locked} = "notlocked";
	    $factoids{lc $1}{forgot} = "";
	    $factoids{lc $1}{unsaved} = 1;
	    foreach(keys %neighborstats) {
		# Tell any neighbors who were wondering.
		if ( $neighborstats{$_}{dunno}{$1} ) {
		    $self->privmsg($_,":INFOBOT:REPLY <$neighborstats{$_}{dunno}{$1}> $1 =is=> $2");
		    print "Telling $_...\n";
		    $neighborstats{$_}{dunno}{$1}="";
		}
	    }
	    $general{mods}++;
	    $general{lastmod}=lc $1;
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid ". $changed);
	}
	else {
	    $self->privmsg(cantalk($room) ? $room : $nick,"Factoid is locked");
	}
    }
    elsif ( $args =~ /^tell ([^ #&]+) about (.+)$/i ) {
	$args =~ s/\\is/is/g;
	my ($target,$what,$prefix);
	$target = $1;
	$what = lc $2;
	$prefix = "$nick wants you to know: ";
	my $fc = $factoids{$what}{content};
	my $date = scalar(localtime);
	$fc =~ s/\$who/$target/g;
	$fc =~ s/\$date/$date/g;
	#print ">> $target -- $fc\n";
	if ( $fc =~ /^<reply>(.+)$/i ) {
	    $self->privmsg($target, $prefix ."$1");
	}
	elsif ( $fc =~ /^<action>(.+)$/i ) {
	    $self->me($target, $prefix ."$1");
	}
	else {
	    $self->privmsg($target, $prefix ."$1 is $fc");
	}
    }
    elsif ( (($args =~ /^(?:what|who|where) (?:is|was) (.+)$/i)) && (exists($factoids{lc $1}{content})) || exists($factoids{lc $args}{content}) ) {
	my $target = cantalk($room) ? $room : $nick;
	my $what = lc ($1 || $args);
	my $prefix = prefix();
	my $fc = $factoids{$what}{content};
	my $date = scalar(localtime);
	$fc =~ s/\$who/$nick/g;
	$fc =~ s/\$date/$date/g;
	#print ">> $target -- $fc\n";
	if ( $fc =~ /^<reply>(.+)$/i ) {
	    $self->privmsg($target, "$1");
	}
	elsif ( $fc =~ /^<action>(.+)$/i ) {
	    $self->me($target, "$1");
	}
	else {
	    $self->privmsg($target, $prefix ."$1 is $fc");
	}
    }
    else {
	my $l=0;
	foreach $fooa ( @commands ) {
	    my @foo = @$fooa;
	    print "Checking for match on module command '$foo[0]'...";
	    if ( $args =~ /$foo[0]/i ) {
		print "found\n";
		print "\tEvaling $foo[1]...";
		my $r = eval($foo[1]);
		print "done\n";
		warn "\tError: $!\n" if !defined $r;
		warn "\tError: $@\n" if $@;
		#warn "\tError\n" if !$r;
		print "\tSending messages...";
		foreach ( split("\n",$r) ) {
		    $self->privmsg(cantalk($room) ? $room : $nick,$_);
		}
		print "done\n";
		$l=1;
		last;
	    }
	    else {
		print "not found\n";
	    }
	}
	if ( $l == 0 ) {
	    foreach (@neighbors) {
		return if $_ eq $nick;
	    }
	    my $target = cantalk($room) ? $room : $nick;
	    unless ( !cantalk($room) && exists $isneighbor{lc $nick} ) {
		$self->privmsg($nick, "'$args': No such factoid or command. I'll ask my neighbors about it. You might need a higher access level. Try HELP.");
	    }
	    my $foo = $args;
	    $foo =~ s/^(?:what|who|where) (?:is|was) (.+)$/$1/;
	    print "Asking my neighbors about $foo"."...\n";
	    foreach ( @neighbors ) {
		print "Asking $_...\n";
		$self->privmsg($_,":INFOBOT:QUERY <$nick> ".lc($foo));
		$neighborstats{lc $_}{waiting}++;
	    }
	    $factoids{lc $foo}{targetwait} = $target;
	}
    }
    #save();
}

# What to do when we receive a private PRIVMSG....
sub on_msg {
    my ($self, $event) = @_;
    my ($nick) = $event->nick;
    my ($args) = ($event->args);

    nblog("overide", "*$nick* $args");
    #$self->privmsg("$nick","'$args'");

    # Infobot Inter-Bot Protocol, or "Asking neighbors for help".
    # See infobot.txt for details.
    if ( $args =~ /:?INFOBOT:QUERY \<([^\>]*)\> (.*)$/ ) {
	$nick = lc $nick;
	$isneighbor{$nick}=1;
	# Another bot is asking us for help.
	if (  exists($factoids{lc $2}{content}) ) { # We know what it is
	    $self->privmsg($nick,":INFOBOT:REPLY <$1> $2 =is=> $factoids{lc $2}{content}");
	}
	else { # We don't know, please let us know when you find out
	    $self->privmsg($nick,":INFOBOT:DUNNO <$1> $2");
	    $factoids{lc $2}{dunno} = 1;
	}
    }
    elsif ( $args =~ /:?INFOBOT:REPLY \<([^\>]*)\> (.*?) =(is|are)=> (.*)$/ && ( $neighborstats{lc $nick}{waiting}>0 || $factoids{lc $2}{dunno} ) ) {
	$nick = lc $nick;
	$isneighbor{$nick}=1;
	# Someone is responding to our dunno or our query. Do the right thing.
	if ( $factoids{lc $2}{dunno} != 1 ) {
	    foreach(keys %neighborstats) {
		# Tell any neighbors who were wondering.
		if ( $neighborstats{$_}{dunno}{$2} ) {
		    $self->privmsg($_,":INFOBOT:REPLY <$neighborstats{$_}{dunno}{$2}> $2 =$3=> $4");
		    $neighborstats{$_}{dunno}{$2}="";
		}
	    }
	    if ( exists($factoids{lc $2}{content}) ) {
		$self->privmsg($1,"$nick knew: $2 $3 $4");
	    }
	    else {			# Tell the room for the FIRST ONE
		$self->privmsg($factoids{lc $2}{targetwait},
			       "$nick knew: $2 $3 $4");
	    }
	}
	if ( !exists($factoids{lc $2}{content}) ) {
	    if ($3 eq "is") {
		$factoids{lc $2}{content} = $4;
	    }
	    else {
		$factoids{lc $2}{content} = "<reply>$2 are $4";
	    }
	    $factoids{lc $2}{creator} = $nick;
	    $factoids{lc $2}{author} = $nick;
	    $factoids{lc $2}{locked} = "notlocked";
	    $factoids{lc $2}{forgot} = "";
	    $factoids{lc $2}{unsaved} = 1;
	}
	$neighborstats{$nick}{waiting}--;
	$factoids{lc $2}{dunno}= 0;
    }
    elsif ( $args =~ /:?INFOBOT:DUNNO \<([^\>]*)\> (.*)$/ && $neighborstats{lc $nick}{waiting}>0 ) {
	$nick = lc $nick;
	$isneighbor{$nick}=1;
	# Someone else doesn't know. Save this and tell them what we find out.
	$neighborstats{$nick}{dunno}{$2}= $1;
    }

    elsif ( $args =~ /^auth (.+)$/i ) {
	open PASSFILE, "data/$nick".".pass";
	my $thepass = <PASSFILE>;
	if ((crypt($1,$thepass) eq $thepass)&&(-f "data/$nick".".pass")) {
	    $userauth{$nick} = $event->from;
	    $self->privmsg($nick, "Password accepted - you are now recognized. Your hostmask is $userauth{$nick}.");
	}
	else {
	    $self->privmsg($nick, "Password Incorrect.");
	    nickserv($self);
	    $self->privmsg("MemoServ", "SEND $nick Someone, prehaps you, but at any rate, using your nickname attempted to identify to nandBot using an incorrect password. Hostmask was ".$event->from." -nandBot");
	}
    }
#    elsif ( $nick =~ /^CCQuote$/i ) {
#	$self->privmsg($nick,randcc());
#    }
    elsif ( $nick =~ /^MSQuote$/i ) {
	$self->privmsg($nick,randms());
    }
    elsif ( $args =~ /^knock knock$/i ) {
	$knock=$nick;
	$self->privmsg($nick,"Who's there?");
    }
    elsif ( $knock eq $nick )  {
	$self->privmsg($nick,$args." who?");
	$knock = "";
    }
    elsif ( $args eq "xyzzy" ) {
	$self->notice($nick,unpack('u',$secret[0]));
	$self->notice($nick,unpack('u',$secret[1]));
	$self->notice($nick,unpack('u',$secret[2]));
	$self->notice($nick,unpack('u',$secret[3]));
	$self->notice($nick,unpack('u',$secret[4]));
	$self->notice($nick,$nick.unpack('u',$secret[5]));
    }
    elsif ( $args =~ /^# GetInfo$/i ) {	# This is the Comic Chat profile thing.
	$self->privmsg($nick,"# HeresInfo: $nbstr");
    }
    elsif ( $args =~ /^# Appears as / ) { # This is the character nandBot uses.
	$self->privmsg($nick,"# Appears as TIKI");
    }
    else {				# It's a nandBot command, maybe?
	$general{commands}++;
	nandbotcmd($self,($event->args),$nick,0,"nandBot","",$event->from);
    }
}

# ...Or a notice.
sub on_notice {
    my ($self, $event) = @_;
    my ($nick) = $event->nick;

    nblog("overide", sprintf("-$nick- %s", ($event->args)));
   # nandbotcmd($self,($event->args),$nick,0,"")
   #   unless $nick =~ /Serv$/i;        # It's a service sending the
					# notice

    if ( $nick =~ /^NickServ$/i && ($event->args)
	 =~ /registered and protected/i ) {
	# NickServ wants my password
	nickserv($self);
    }
}

# What to do when we receive channel text.
sub on_public {
    my ($self, $event) = @_;
    my @to = $event->to;
    my ($nick, $mynick) = ($event->nick, $self->nick);
    my ($arg) = ($event->args);
    my $wap = 0;
    my $comic = 0;

    $general{messages}++;
    #[04:10:53P] (nand2): # Appears as ANNA
    #[04:11:01P] (Ryan): oi
    #[04:11:03P] (nand2): (#G010E010M1) This is a message from comic chat
    #[04:11:04P] (Ryan): MSChat?

    # Special handling for Comic Chat users (bleh) (Character definition)
    if ( $arg =~ /^# Appears as / ) {
	$self->privmsg($nick,"# Appears as TIKI");
    }
    # Special handling for WAP users
    if ( $arg =~ /^\<(.+)\> (.+)$/i && cantalk(@to)) {
	$general{via}++;
	$relaynick = $nick;
	$arg = $2;
	$nick = $1;
	$wap = 1;
    }
    # Special handling for Comic Chat users (bleh)
    if ( $arg =~ /^\(\#[^\)]+\) (.+)$/i ) {
	$arg = $1;
	$comic = 1;
    }

    #[04:11:58P] (Ryan): eww
    #[04:11:58P] (Ryan): lol
    #[04:12:01P] (Ryan): i hate comic chat
    #[04:12:06P] (Hemi[Linux]): lol
    #[04:12:06P] (nand2): (#G410E?10M1) so do I

    # Note that $event->to() returns a list (or arrayref, in scalar
    # context) of the message's recipients, since there can easily be
    # more than one.
    my $cmd = $arg;
    $cmd =~ s/[\?\!\.\,\/\;]*$//;	# Remove punctuation
    unless ( $botoff ) {
	$seenroom = $to[0];
	if ( $wap ) {
	    $seenroom .= " (via the $relaynick robot)";
	}
	$users{lc $nick}{seentxt} = $arg;
	$users{lc $nick}{seenroom} = $seenroom;
	$users{lc $nick}{seentime} = time;
	my $pointsto=$users{lc $nick}{doredir} ? $users{lc $nick}{redirect} : $nick;
	$users{lc $pointsto}{points} ++;
	$users{lc $nick}{unsaved} = 1;
	$users{lc $pointsto}{unsaved} = 1;
	if ( $general{maxpoints} < $users{lc $pointsto}{points} ) {
	    $general{maxpoints} = $users{lc $pointsto}{points};
	    $general{maxpointsnick} = $pointsto;
	}
	if ( $arg =~ /^(grr+)/i && ($users{lc $nick}{angeroptout}||0)<2&& time > $general{angertime}+30) {
	    $self->privmsg( @to, "Your anger ratio is ". length($1) .".") if cantalk(@to);
	    if ( ($users{lc $nick}{angeroptout} || 0) < 1 ) {
		$self->privmsg( $nick,"This is a one-time message only. You can opt-out of the 'Anger Ratio' feature by whispering to me: 'OPTOUT ANGER'");
		$users{lc $nick}{angeroptout}=1;
	    }
	    $general{angertime}=time;
	}
	if ( $arg =~ /nandy/i && 0) {
	    $self->privmsg( $nick, "I'VE TOLD YOU ONCE IF I'VE TOLD YOU A HUNDRED TIMES, YOU DO NOT SAY 'nandy' IN - FRONT - OF - ME!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
	}
	elsif ( $arg =~ /(^|\s)([^ \t\(\)]+)\(sp\??\)(\s|$)/i ) {
	    my $dictWord = $2;
	    if ( \&Speller ) {
		$self->privmsg($nick,Speller($dictWord));
	    }
	}
	if ($cmd =~ /^$mynick$/i && cantalk($to[0])) {
	    $self->privmsg($to[0],"Yes, ". (($nick eq "nand" && 0)? "Master" : $nick) ."?");
	}
    }

    if ($cmd =~ /^$mynick[\-\,\:]? (.+)/i || $cmd =~ /^~(.+)/) {
	$general{commands}++;
	nblog("$channel".' [OR]', "<$to[0]: $nick> $arg\n");
	nandbotcmd($self,$1,$nick,1,$event->to,$wap,$event->from);
    }
    else {
	nblog($to[0], "<$to[0]: $nick> $arg\n");
    }
    # $self->privmsg("$to[0]","$1$arg") if $arg =~ /(^\w+) ?.*$/;
    save() if $mc++ == 25;		# Save every 25 messages.
    #print $mc;
}

# Prints the names of people in a channel when we enter.
sub on_names {
    my ($self, $event) = @_;
    my (@list, $channel) = ($event->args);    # eat yer heart out, mjd!

    # splice() only works on real arrays. Sigh.
    ($channel, @list) = splice @list, 2;

    print "Users on $channel: @list\n";
}

# Replys with a VERSION statement
sub on_version {
    my ($self, $event) = @_;
    my $nick = $event->nick;

    $self->ctcp_reply($nick, 'VERSION '.$nbstr);
    # $self->ctcp_reply($nick, sprintf("VERSION Net::IRC module on Perl %vd on RedHat Linux 7.2",$^V));
    print "*** CTCP VERSION request from $nick received\n";
}

sub on_ping {				# Yells about incoming CTCP PINGs.
    my ($self, $event) = @_;
    my $nick = $event->nick;

    $self->ctcp_reply($nick, "PING", join (' ', $event->args));
    print "*** CTCP PING request from $nick received\n";
}

# *sigh*
# <- :tnandBot!nandBot@FC-20363406.dsl.mindspring.com NOTICE nand :TIME
sub on_time {				# Returns the current time.
    my ($self, $event) = @_;
    my $nick = $event->nick;
                                 # Why can't I make this two params like ping?
    $self->ctcp_reply($nick, "TIME ".scalar(localtime));
    print "*** CTCP TIME request from $nick received\n";
}

# [04:07:05P] (Kandy): you're staring at soccer chicks?

sub on_finger {				# CTCP Finger returns the nandBot url
    my ($self, $event) = @_;
    my $nick = $event->nick;
                                   # Of course VERSION does it this way too.
    $self->ctcp_reply($nick, "FINGER ".$nbstr);
    print "*** CTCP FINGER request from $nick received\n";
}

sub on_userinfo {			# As does CTCP userinfo
    my ($self, $event) = @_;
    my $nick = $event->nick;

    $self->ctcp_reply($nick, "USERINFO ".$nbstr);
    print "*** CTCP USERINFO request from $nick received\n";
}
# Gives lag results for outgoing PINGs.
sub on_ping_reply {
    my ($self, $event) = @_;
    my ($args) = ($event->args)[1];
    my ($nick) = $event->nick;

    $args = time - $args;
    print "*** CTCP PING reply from $nick: $args sec.\n";
    if ( $general{pingwaiting} ) {
	$self->privmsg($general{pingwaiting},"Lag: $args sec.");
	$general{pingwaiting}="";
    }
}

# Change our nick if someone stole it.
sub on_nick_taken {
    my ($self) = shift;

    $self->nick(substr($self->nick, -1) . substr($self->nick, 0, 8));
}

# Display formatted CTCP ACTIONs.
sub on_action {
    my ($self, $event) = @_;
    my ($nick, $mynick, @args) = ($event->nick, $self->nick, $event->args);
    @to = $event->to;

    nblog($to[0], "$to[0]: *"."$nick @args\n");
    $sr = join( " ",@args);		# Silliness Replys ;)
    if($sr =~ /^pokes nandBot/i) {
	$self->me(cantalk($to[0]) ? $to[0] : $nick,"is here");
	print "nandBot is here.\n";
	#$self->ctcp($mynick,"ping ".time);
	#$general{pingwaiting} = (cantalk($to[0]) ? $to[0] : $nick);
    }
    if ( cantalk($to[0]) ) {
	my $rm = $to[0];
	if ($sr =~ /^gives $mynick a (bot )?(biscuit|cookie)/i ) {
	    # [14:20:39] <nand> Am I over-feeding nandBot or something?
	    $self->privmsg($rm,"*munch* Thanks");
	}
	elsif ($sr =~ /^kicks $mynick/i ) { # Hee Hee Hee
	    $self->me($rm, "kicks $nick back");
	}
    }
    #print "'".join(" ",@args)."'\n"
}

#Handles quiting - badly
sub quit {
    # print "Saving settings...\n";
    save();
    if ( $_[0] ) {
	my $self = $_[0];
	$self->quit($nbstr);
    }
    unlink "nandbot.pid";
    exit(0);
}

# Reconnect to the server when we die.
sub on_disconnect {
    my ($self, $event) = @_;

    nblog("overide","Disconnected from ". $event->from() . " (".
      ($event->args())[0] ."). Attempting to reconnect...\n");
    my $done = 0;
    while(!$done ) {
	$done = $self->connect();
    }
}

# Look at the topic for a channel you join.
sub on_topic {
    my ($self, $event) = @_;
    my @args = $event->args();

    # Note the use of the same handler sub for different events.

    if ($event->type() eq 'notopic') {
	print "No topic set for $args[1].\n";
	# If it's being done _to_ the channel, it's a topic change.
    } elsif ($event->type() eq 'topic' and $event->to()) {
	nblog($event->to,"Topic change for ". $event->to() .": $args[0]\n");
    } else {
	print "The topic for $args[1] is \"$args[2]\".\n";
    }
}


sub pickrandom {   # Choose a random word from the @words array.
    return $words[ rand scalar @words ];
}
sub ateball {      # Choose a random word from the @ateball array.
    return $ateball[ rand scalar @ateball ];
}
sub prefix {       # Choose a random word from the @prefixes array.
    return $prefixes[ rand scalar @prefixes ];
}
sub randcc {       # Choose a random word from the @ccquotes array.
    return $ccquotes[ rand scalar @ccquotes ];
}
sub randms {       # Choose a random word from the @msquotes array.
    return $msquotes[ rand scalar @msquotes ];
}


print "Installing handler routines...";

$conn->add_handler('cping',  		\&on_ping);
$conn->add_handler('crping', 		\&on_ping_reply);
$conn->add_handler('cversion',  	\&on_version);
$conn->add_handler('msg',    		\&on_msg);
$conn->add_handler('notice', 	  	\&on_notice);
#$conn->add_handler('chat', 		\&on_chat);
$conn->add_handler('public', 		\&on_public);
$conn->add_handler('caction', 		\&on_action);
$conn->add_handler('join',  		\&on_join);
$conn->add_handler('part',  		\&on_part);
#$conn->add_handler('cdcc',  		\&on_dcc);
$conn->add_handler('topic', 		\&on_topic);
$conn->add_handler('notopic',		\&on_topic);
$conn->add_handler('kick',		\&on_kick);
$conn->add_handler('mode',		\&on_mode);
$conn->add_handler('invite',		\&on_invite);
#$conn->add_handler('crversion',	\&on_versionreply);
$conn->add_handler('ctime',		\&on_time);
$conn->add_handler('cfinger',		\&on_finger);
$conn->add_handler('cuserinfo',		\&on_userinfo);

$conn->add_global_handler([ 251,252,253,254,302,255 ], \&on_init);
$conn->add_global_handler('disconnect',	\&on_disconnect);
$conn->add_global_handler(376,		\&on_connect);
$conn->add_global_handler(433,		\&on_nick_taken);
$conn->add_global_handler(353,		\&on_names);

print "done\n";

# Load all modules
print "Loading modules...";
opendir LIBDIR, "lib/";
my @libdir=readdir LIBDIR;
@libdir = sort @libdir;
foreach (@libdir) {
    if ( $_ =~ /\.pl$/i ) {
	my $name = $_;
	$name =~ s/\.pl$//i;
	print "$name...";
	my $r = do "lib/$_";
	warn "\nError loading $name: $!\n" if !defined $r;
	warn "\nError loading $name: $@\n" if $@;
	warn "\nError loading $name\n" if !$r;
    }
}
print "done\n";

#print "Saving PID...";
#system "echo $$ > nandbot.pid";
open PF, ">nandbot.pid";
print PF $$;
close PF;
#print " done.\n";
$SIG{HUP}=\&quit;
$SIG{TERM}=\&quit;
$SIG{KILL}=\&quit;

$general{startup} = time;
$general{commands} = 0;
$general{messages} = 0;
$general{via} = 0;
$general{mods} = 0;
$general{lastmod} = "";
print "starting...\n";
$irc->start;
BEGIN {print "done\n" }
